﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;

namespace RadiativeAC
{
    class Program
    {

        #region 定数

        const double CPA = 1.006;

        #endregion

        #region 変数

        static bool isStatic;
        static double tStep, endTime, tSol;
        static double sWU, sWL, sH, sFU, sFL, sPU, sPL;
        static double alpha_C_H, alpha_C_PU, alpha_C_PL, alpha_C_WU, alpha_C_WL, alpha_C_FU, alpha_C_FL;
        static double cap_WU, cap_WL, cap_H, cap_F, cap_RU, cap_RL;
        static double r_WUO, r_WUI, r_WLO, r_WLI, r_H, r_F;
        static double zoneVentilation;
        static double qH;
        static Matrix alpha_R_XX;

        #endregion

        static void Main(string[] args)
        {
            //シミュレーション条件読み込み
            readParameters();

            //境界条件読み込み
            string[] title;
            double[][] variableValues;
            int[][] variableIndices;
            int[][] boundaryIndices;
            readBoundaries(out title, out variableValues, out variableIndices, out boundaryIndices);

            //計算実行+書き出し
            double[][] outputs;
            using (StreamWriter sWriter = new StreamWriter("output.csv",false, Encoding.GetEncoding("Shift_JIS")))
            {
                //タイトル行
                sWriter.WriteLine("TWCU,TWCL,THC,TFC,TRU,TRL,THS,TP,TFS,TWSU,TWSL,QP,QRU,QRL");

                for (int i = 0; i < variableValues.Length; i++)
                {
                    execute(variableValues[i], variableIndices[i], boundaryIndices[i], out outputs);
                    sWriter.WriteLine(title[i]);
                    for (int j = 0; j < outputs.Length; j++)
                    {
                        for (int k = 0; k < outputs[j].Length; k++)
                        {
                            sWriter.Write(outputs[j][k] + ",");
                        }
                        sWriter.WriteLine();
                    }
                }

                //放射熱伝達率書き出し
                sWriter.WriteLine("放射熱伝達率");
                for (uint i = 0; i < alpha_R_XX.Rows; i++)
                {
                    for (uint j = 0; j < alpha_R_XX.Columns; j++)
                    {
                        sWriter.Write(alpha_R_XX.GetValue(i,j) + ",");
                    }
                    sWriter.WriteLine();
                }
            }
        }


        #region 放射熱交換係数の計算処理

        /// <summary>形態係数に基いて放射熱伝達率を計算する</summary>
        /// <param name="f">形態係数行列</param>
        /// <param name="e">放射率ベクトル</param>
        /// <param name="mat_ar">放射熱伝達率arbWnn[W/m2K]</param>
        private static void radex(double[,] f, double[] e, out Matrix mat_ar)
        {
            double SS = 4.0 * 5.67e-8 * Math.Pow(20.0d + 273.15, 3);

            uint num = (uint)f.GetLength(0);
            Matrix mat_e = new Matrix(num, num);
            Matrix mat_f = new Matrix(num, num);
            Matrix mat_einv = new Matrix(num, num);
            Matrix mat_phi = new Matrix(num, num);
            mat_ar = new Matrix(num, num);
            for (uint i = 0; i < num; i++)
            {
                for (uint j = 0; j < num; j++)
                {
                    mat_e.SetValue(i, j, -f[i, j] * (1.0 - e[j]) / e[j]);
                    mat_f.SetValue(i, j, -f[i, j]);
                }
                mat_e.AddValue(i, i, 1.0 / e[i]);
                mat_f.AddValue(i, i, 1.0);
            }
            mat_e.GetInverse(ref mat_einv);
            mat_einv.MatrixProduct(mat_f, ref mat_phi, 1, 0);

            //放射熱伝達率arbWnn[W/m2K]の計算
            for (uint i = 0; i < num; i++)
            {
                for (uint j = 0; j < num; j++)
                {
                    mat_ar.SetValue(i, j, Math.Abs(mat_phi.GetValue(i, j) * SS));
                }
            }
        }

        #endregion

        #region シミュレーション条件読み込み処理

        private static void readParameters()
        {
            using (StreamReader sReader = new StreamReader("parameter.csv", Encoding.GetEncoding("Shift_JIS")))
            {
                //境界条件ファイルのデータ区切規則
                string[] delimiter = new string[] { ",", "//", "\t" };

                //計算時間間隔[sec]
                tStep = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                isStatic = (tStep <= 0);

                //計算終了時間[sec]
                endTime = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);

                //室内の物体の放射率
                double[] epsiron = new double[5];    //人体、パネル、上部壁体、下部壁体、内壁・家具の放射率
                epsiron[0] = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                epsiron[1] = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                epsiron[2] = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                epsiron[3] = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                epsiron[4] = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);

                //相当外気温度[℃]
                tSol = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                
                //表面積読み込み処理
                //人体表面積[㎡]
                sH = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //放射パネル上部表面積[㎡]
                sPU = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //放射パネル下部表面積[㎡]
                sPL = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //上部壁体表面積[㎡]
                sWU = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //下部壁体表面積[㎡]
                sWL = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //家具上部表面積[㎡]
                sFU = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //家具下部表面積[㎡]
                sFL = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);

                //対流熱伝達率読み込み処理
                //人体表面対流熱伝達率[W/㎡K]
                alpha_C_H = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //放射パネル表面対流熱伝達率（上部）[W/㎡K]
                alpha_C_PU = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //放射パネル表面対流熱伝達率（下部）[W/㎡K]
                alpha_C_PL = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //上部壁体対流熱伝達率[W/㎡K]
                alpha_C_WU = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //下部壁体対流熱伝達率[W/㎡K]
                alpha_C_WL = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //内壁・家具の対流熱伝達率（上部）[W/㎡K]
                alpha_C_FU = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //内壁・家具の対流熱伝達率（下部）[W/㎡K]
                alpha_C_FL = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);

                //人体情報読み込み処理
                //人体発熱[W]
                qH = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //人体の熱容量[J/K]
                cap_H = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //人体の熱抵抗[K/W]
                r_H = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);

                //上部壁体情報読み込み処理
                //上部壁体熱容量[J/K]
                cap_WU = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //上部壁体屋内側熱抵抗[K/W]
                r_WUI = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //上部壁体屋外側熱抵抗[K/W]
                r_WUO = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);

                //下部壁体情報読み込み処理
                //下部壁体熱容量[J/K]
                cap_WL = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //下部壁体屋内側熱抵抗[K/W]
                r_WLI = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //下部壁体屋外側熱抵抗[K/W]
                r_WLO = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);

                //家具情報読み込み処理
                //家具の熱容量[J/K]
                cap_F = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //家具の熱抵抗[K/W]
                r_F = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);

                //ゾーン情報読み込み処理
                //上部ゾーン熱容量[J/K]
                cap_RU = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //下部ゾーン熱容量[J/K]
                cap_RL = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                //ゾーン間換気[CMH]→[g/s]
                zoneVentilation = double.Parse(sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries)[1]);
                zoneVentilation = zoneVentilation * 1.2 * 1000 / 3600d;

                //形態係数読み込み処理
                double[,] ff = new double[5,5];
                for (int i = 0; i < ff.GetLength(0); i++)
                {
                    string[] fBuff = sReader.ReadLine().Split(delimiter, StringSplitOptions.RemoveEmptyEntries);
                    for (int j = 0; j < ff.GetLength(1); j++)
                    {
                        ff[i, j] = double.Parse(fBuff[j + 1]);
                    }
                }

                //放射熱交換係数を計算する
                radex(ff, epsiron, out alpha_R_XX);
            }
        }

        #endregion

        #region 境界条件読み込み処理

        private static void readBoundaries(out string[] title, out double[][] variableValues, out int[][] variableIndices, out int[][] boundaryIndices)
        {
            using (StreamReader sReader = new StreamReader("boundary.csv", Encoding.GetEncoding("Shift_JIS")))
            {
                //境界条件ファイルのデータ区切規則
                string[] delimiter = new string[] { ",", "//", "\t" };

                //ヘッダ行読み飛ばし
                sReader.ReadLine();

                string buff;
                List<double[]> vValues = new List<double[]>();
                List<int[]> vIndices = new List<int[]>();
                List<int[]> bIndices = new List<int[]>();
                List<string> titl = new List<string>();
                while ((buff = sReader.ReadLine()) != null)
                {
                    string[] sbf = buff.Split(delimiter, StringSplitOptions.RemoveEmptyEntries);
                    
                    //計算条件読み込み処理
                    titl.Add(sbf[0]);

                    //状態変数読み込み処理
                    double[] vv = new double[14];
                    for (int i = 0; i < vv.Length; i++) vv[i] = double.Parse(sbf[i + 1]);
                    vValues.Add(vv);

                    //境界条件番号読み込み処理
                    int[] vi = new int[11];
                    int[] bi = new int[3];
                    int viIndex = 0;
                    int biIndex = 0;
                    for (int i = 0; i < 14; i++)
                    {
                        if (bool.Parse(sbf[i + 15]) == true)
                        {
                            bi[biIndex] = i;
                            biIndex++;
                        }
                        else
                        {
                            vi[viIndex] = i;
                            viIndex++;
                        }
                    }
                    vIndices.Add(vi);
                    bIndices.Add(bi);
                }

                title = titl.ToArray();
                variableValues = vValues.ToArray();
                variableIndices = vIndices.ToArray();
                boundaryIndices = bIndices.ToArray();
                
            }
        }

        #endregion

        #region シミュレーション実行処理

        private static void execute(double[] variableValues, int[] variableIndices, int[] boundaryIndices, out double[][] outputs)
        {
            //静的計算の場合にはタイムステップを馬鹿みたいに大きくとっておく（0割判定面倒くさいので、これでも答えは一緒）
            int iterNum;
            if (isStatic)
            {
                iterNum = 1;
                tStep = 999999999;
                outputs = new double[1][];
            }
            else
            {
                iterNum = (int)(endTime / tStep);
                outputs = new double[iterNum][];
            }

            //A行列作成
            double[,] ax = new double[11, 14];
            ax[0, 0] = -(cap_WU / tStep + 1 / r_WUO + 1 / r_WUI);
            ax[0, 9] = 1 / r_WUI;
            ax[1, 1] = -(cap_WL / tStep + 1 / r_WLO + 1 / r_WLI);
            ax[1, 10] = 1 / r_WLI;
            ax[2, 2] = -(cap_H / tStep + 1 / r_H);
            ax[2, 6] = 1 / r_H;
            ax[3, 3] = -(cap_F / tStep + 1 / r_F);
            ax[3, 8] = 1 / r_F;
            ax[4, 4] = -(cap_RU / tStep + zoneVentilation * CPA + sPU * alpha_C_PU + sFU * alpha_C_FU + sWU * alpha_C_WU);
            ax[4, 5] = zoneVentilation * CPA;
            ax[4, 7] = sPU * alpha_C_PU;
            ax[4, 8] = sFU * alpha_C_FU;
            ax[4, 9] = sWU * alpha_C_WU;
            ax[4, 12] = 1;
            ax[5, 4] = zoneVentilation * CPA;
            ax[5, 5] = -(cap_RL / tStep + zoneVentilation * CPA + sPL * alpha_C_PL + sFL * alpha_C_FL + sWL * alpha_C_WL + sH * alpha_C_H);
            ax[5, 6] = sH * alpha_C_H;
            ax[5, 7] = sPL * alpha_C_PL;
            ax[5, 8] = sFL * alpha_C_FL;
            ax[5, 10] = sWL * alpha_C_WL;
            ax[5, 13] = 1;
            ax[6, 2] = 1d / r_H;
            ax[6, 5] = sH * alpha_C_H;
            ax[6, 6] = -(1 / r_H + sH * (alpha_C_H + alpha_R_XX.GetValue(0, 1) + alpha_R_XX.GetValue(0, 4) + alpha_R_XX.GetValue(0, 2) + alpha_R_XX.GetValue(0, 3)));
            ax[6, 7] = sH * alpha_R_XX.GetValue(0, 1);
            ax[6, 8] = sH * alpha_R_XX.GetValue(0, 4);
            ax[6, 9] = sH * alpha_R_XX.GetValue(0, 2);
            ax[6, 10] = sH * alpha_R_XX.GetValue(0, 3);
            ax[7, 4] = sPU * alpha_C_PU;
            ax[7, 5] = sPL * alpha_C_PL;
            ax[7, 6] = (sPU + sPL) * alpha_R_XX.GetValue(1, 0);
            ax[7, 7] = -(sPU * alpha_C_PU + sPL * alpha_C_PL + (sPU + sPL) * (alpha_R_XX.GetValue(1, 0) + alpha_R_XX.GetValue(1, 4) + alpha_R_XX.GetValue(1, 2) + alpha_R_XX.GetValue(1, 3)));
            ax[7, 8] = (sPU + sPL) * alpha_R_XX.GetValue(1, 4);
            ax[7, 9] = (sPU + sPL) * alpha_R_XX.GetValue(1, 2);
            ax[7, 10] = (sPU + sPL) * alpha_R_XX.GetValue(1, 3);
            ax[7, 11] = 1;
            ax[8, 3] = 1 / r_F;
            ax[8, 4] = sFU * alpha_C_FU;
            ax[8, 5] = sFL * alpha_C_FL;
            ax[8, 6] = (sFU + sFL) * alpha_R_XX.GetValue(4, 0);
            ax[8, 7] = (sFU + sFL) * alpha_R_XX.GetValue(4, 1);
            ax[8, 8] = -(sFU * alpha_C_FU + sFL * alpha_C_FL + (sFU + sFL) * (alpha_R_XX.GetValue(4, 0) + alpha_R_XX.GetValue(4, 1) + alpha_R_XX.GetValue(4, 2) + alpha_R_XX.GetValue(4, 3)) + 1 / r_F);
            ax[8, 9] = (sFU + sFL) * alpha_R_XX.GetValue(4, 2);
            ax[8, 10] = (sFU + sFL) * alpha_R_XX.GetValue(4, 3);
            ax[9, 0] = 1 / r_WUI;
            ax[9, 4] = sWU * alpha_C_WU;
            ax[9, 6] = sWU * alpha_R_XX.GetValue(2, 0);
            ax[9, 7] = sWU * alpha_R_XX.GetValue(2, 1);
            ax[9, 8] = sWU * alpha_R_XX.GetValue(2, 4);
            ax[9, 9] = -sWU * (alpha_C_WU + alpha_R_XX.GetValue(2, 0) + alpha_R_XX.GetValue(2, 1) + alpha_R_XX.GetValue(2, 3) + alpha_R_XX.GetValue(2, 4)) - 1 / r_WUI;
            ax[9, 10] = sWU * alpha_R_XX.GetValue(2, 3);
            ax[10, 1] = 1 / r_WLI;
            ax[10, 5] = sWL * alpha_C_WL;
            ax[10, 6] = sWL * alpha_R_XX.GetValue(3, 0);
            ax[10, 7] = sWL * alpha_R_XX.GetValue(3, 1);
            ax[10, 8] = sWL * alpha_R_XX.GetValue(3, 4);
            ax[10, 9] = sWL * alpha_R_XX.GetValue(3, 2);
            ax[10, 10] = -sWL * (alpha_C_WL + alpha_R_XX.GetValue(3, 0) + alpha_R_XX.GetValue(3, 1) + alpha_R_XX.GetValue(3, 4) + alpha_R_XX.GetValue(3, 2)) - 1 / r_WLI;

            for (int num = 0; num < iterNum; num++)
            {
                Vector vectorB;
                Matrix matrixC;
                makeBCMatrix(variableValues, variableIndices, boundaryIndices, ax, out vectorB, out matrixC);

                Matrix matrixXC = new Matrix(11, 11);
                Vector vectorD = new Vector(11);
                matrixC.GetInverse(ref matrixXC);
                matrixXC.VectorProduct(vectorB, ref vectorD, 1, 0);
                outputs[num] = new double[14];
                for (uint i = 0; i < vectorD.Size; i++)
                {
                    outputs[num][variableIndices[i]] = variableValues[variableIndices[i]] = vectorD.GetValue(i);                     
                }
                for (int i = 0; i < 3; i++) outputs[num][boundaryIndices[i]] = variableValues[boundaryIndices[i]];
            }
        }

        private static void makeBCMatrix(double[] variableValues, int[] variableIndices, int[] boundaryIndices, double[,] matrixA, out Vector vectorB, out Matrix matrixC)
        {
            vectorB = new Vector(11);
            matrixC = new Matrix(11, 11);
            
            vectorB.SetValue(0, -cap_WU * variableValues[0] / tStep - tSol / r_WUO);
            vectorB.SetValue(1, -cap_WL * variableValues[1] / tStep - tSol / r_WLO);
            vectorB.SetValue(2, -cap_H * variableValues[2] / tStep - qH);
            vectorB.SetValue(3, -cap_F * variableValues[3] / tStep);
            vectorB.SetValue(4, -cap_RU * variableValues[4] / tStep);
            vectorB.SetValue(5, -cap_RL * variableValues[5] / tStep);

            for (uint i = 0; i < matrixA.GetLength(0); i++)
            {
                //Bベクトル作成処理
                for (int j = 0; j < 3; j++)
                {
                    vectorB.AddValue(i, -matrixA[i, boundaryIndices[j]] * variableValues[boundaryIndices[j]]);
                }

                //C行列作成処理
                for (uint j = 0; j < 11; j++)
                {
                    matrixC.SetValue(i, j, matrixA[i, variableIndices[j]]);
                }
            }
        }

        #endregion

    }
}
